//! When you use the create_tests flag on OptimizationSettings,
//! simplified test cases are created for each optimization step
//! and appended to this file

#![feature(test)]
#![allow(unused_imports)]
#![allow(unused_variables)]

use rust_circuit::all_imports::*;
use rust_circuit::circuit::debugging::assert_is_close;

// Start of machine generated tests
